/* rpm2tar.c: convert the files contained in an RPM package into a gzipped
              tar archive, preserving ownership and permissions.
 * created 1999-Mar-26 17:10 jmk
 * autodate: 2000-Nov-27 00:58
 * 
 * by Jim Knoble <jmknoble@jmknoble.cx>
 * Copyright (C) 1999,2000 Jim Knoble
 * 
 * Disclaimer:
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * express or implied, including but not limited to the warranties of
 * merchantability, fitness for a particular purpose and
 * noninfringement. In no event shall the author(s) be liable for any
 * claim, damages or other liability, whether in an action of
 * contract, tort or otherwise, arising from, out of or in connection
 * with the software or the use or other dealings in the software.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <rpmlib.h>

#define BUF_LEN	4096

#ifndef DEBUG_STUFF
#define DEBUG_STUFF 0
#endif

static char *prog_name;

static char *bad_dirs[] =  {
   ".",
   "..",
   "/",
   "//",
   "/bin",
   "/bin/",
   "/boot",
   "/boot/",
   "/dev",
   "/dev/",
   "/etc",
   "/etc/",
   "/home",
   "/home/",
   "/lib",
   "/lib/",
   "/lost+found",
   "/lost+found/",
   "/mnt",
   "/mnt/",
   "/net",
   "/net/",
   "/opt",
   "/opt/",
   "/proc",
   "/proc/",
   "/root",
   "/root/",
   "/sbin",
   "/sbin/",
   "/tmp",
   "/tmp/",
   "/usr",
   "/usr/",
   "/usr/bin",
   "/usr/bin/",
   "/usr/lib",
   "/usr/lib/",
   "/usr/local",
   "/usr/local/",
   "/usr/sbin",
   "/usr/sbin/",
   "/var",
   "/var/",
   "/var/log",
   "/var/log/",
   NULL
};

static char *usage_text =
   "Usage: %s <rpm-package> [<working-dir> [<output-file>]]\n"
   ;

static char *help_text =
   "\n"
   "Create a gzipped tar archive of the files contained in <rpm-package>.\n"
   "Use <working-dir> as a temporary working directory (defaults to\n"
   "'%s').\n"
   "\n"
   "The resulting tar archive is called <output-file> (defaults to\n"
   "<name>-<version>-<release>.<arch>.tar.gz).\n"
   "\n"
   "You should be root to use this program.\n"
   ;

void print_help(char *prog_name, char *work_dir) {
   fprintf(stderr, usage_text, prog_name);
   fprintf(stderr, help_text, work_dir);
   exit(1);
}

void do_command(char *buffer) {
   int status;
   
   fprintf(stderr, "+ %s\n", buffer);
   status = system(buffer); 
   if (0 != status) {
      fprintf(stderr,
	      "%s: fatal error: The command '%s' failed.\n",
	      prog_name, buffer);
      exit(1);
   }
}

int main (int argc, char **argv) {
   char *slash;
   char *work_dir = "/tmp/rpm2tar";
   char *outfilename = NULL;
   char *pkg_filename = NULL;
   FD_t pkg_handle = NULL;
   
   struct stat statinfo;
   
   Header pkg_hdr;
   int pkg_is_source;
   int pkg_type_major;
   int pkg_type_minor;
   char *entry;
   int_32 entry_type;
   int_32 entry_count;
   
   char buffer[BUF_LEN];
   char cwd[BUF_LEN/2];
   
   char *pkg_name;
   char *pkg_version;
   char *pkg_release;
   char *pkg_arch;
   char **pkg_files;
   int_16 *pkg_modes;
   char **pkg_owners;
   char **pkg_groups;
   char *f;
   char *owner;
   char *group;
   int n, i;
   int l = 0;
   
   char *cpio_cmd = "cpio -idm ";
   char *tar_cmd = "tar -cvvf - * |gzip -9 -c";
   char *tar_suffix = ".tar.gz";
   
   int status;
   
   prog_name = argv[0];
   slash = strrchr(prog_name, '/');
   if (NULL != slash) {
      prog_name = ++slash;
   }
   
   if ((argc >= 2) &&
       ((0 != strcmp(argv[1], "--help")) &&
	(0 != strcmp(argv[1], "-h")) &&
	(0 != strcmp(argv[1], "-help")) &&
	(0 != strcmp(argv[1], "-?")))) {
      pkg_filename = argv[1];
   } else {
      print_help(prog_name, work_dir);
   }
   if (argc >= 3) {
      work_dir = argv[2];
   }
   for (i = 0; NULL != bad_dirs[i]; i++) {
      if (0 == strcmp(bad_dirs[i], work_dir)) {
	 fprintf(stderr,
		 "Aaahh! Don't use '%s' as a working directory!\n"
		 "The working directory is deleted as part of normal "
		 "operations!\n", work_dir);
	 exit(1);
      }
   }
   status = lstat(work_dir, &statinfo);
   if ((0 == status) && (!(S_ISDIR(statinfo.st_mode)))) {
      fprintf(stderr,
	      "Caution! Your working directory '%s' already exists,\n"
	      "but it's not a directory.  You may be the victim of a symlink\n"
	      "attack (or you may just be forgetful).  Please either use a\n"
	      "different working directory or remove '%s' before\n"
	      "continuing.\n", work_dir, work_dir);
      exit(1);
   }
   
   if (argc >= 4) {
      outfilename = argv[3];
   }
   
   /* Sanity checking. */
   if ((NULL == pkg_filename) || (0 == strcmp(pkg_filename, ""))) {
      fprintf(stderr,
	      "Hmm ... you seem to have specified an empty RPM package filename.\n"
	      "Unfortunately, i don't know how to handle that.  Sorry.\n"
	      "*cough* *choke* *sputter* *fizzle*\n");
      exit(1);
   } else if (0 == strcmp(pkg_filename, "-")) {
      fprintf(stderr,
	      "Umm ... hmm.  You seem to have asked me to read the RPM package\n"
	      "from the standard input.  Unfortunately, i don't know how to do that.\n"
	      "Sorry.\n"
	      "*cough* *choke* *sputter* *fizzle*\n");
      exit(1);
   }

   /* Get info from RPM package */
   pkg_handle = Fopen(pkg_filename, "r.ufdio");
   if (NULL == pkg_handle || Ferror(pkg_handle)) {
      fprintf(stderr, "%s: error: %s: %s\n", prog_name,
	      pkg_filename, Fstrerror(pkg_handle));
      exit(1);
   }
   status = rpmReadPackageHeader(pkg_handle, &pkg_hdr, &pkg_is_source,
				 &pkg_type_major, &pkg_type_minor);
   switch (status) {
    case 0:
      Fclose(pkg_handle);
      break;
    case 1:
      fprintf(stderr, "%s: error: '%s' doesn't appear to be an RPM package.\n",
	      prog_name, pkg_filename);
      perror("This error message may be relevant");
      exit(1);
      break;
    case 2:
      fprintf(stderr,
	      "%s: error: I seem to have encountered an error reading '%s'.\n",
	      prog_name, pkg_filename);
      perror("This error message may be relevant");
      exit(1);
      break;
    default:
      fprintf(stderr,
	      "%s: error: This is really weird; rpmReadPackageHeader() "
	      "isn't supposed to return %d.\n", prog_name, status);
      exit(1);
   }
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_NAME, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the package name.\n",
	      prog_name);
      exit(1);
   }
   pkg_name = (char *)entry;
#if DEBUG_STUFF
   fprintf(stderr, "Name: %s\n", pkg_name);
#endif
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_VERSION, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the package version.\n",
	      prog_name);
      exit(1);
   }
   pkg_version = (char *)entry;
#if DEBUG_STUFF
   fprintf(stderr, "Version: %s\n", pkg_version);
#endif
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_RELEASE, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the package release.\n",
	      prog_name);
      exit(1);
   }
   pkg_release = (char *)entry;
#if DEBUG_STUFF
   fprintf(stderr, "Release: %s\n", pkg_release);
#endif
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_ARCH, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the package architecture.\n",
	      prog_name);
      exit(1);
   }
   pkg_arch = (char *)entry;
#if DEBUG_STUFF
   fprintf(stderr, "Arch: %s\n", pkg_arch);
#endif
   
   /* Build output filename */
   if (NULL == getcwd(cwd, BUF_LEN)) {
      fprintf(stderr,
	      "%s: error: I don't seem to have enough space to hold the "
	      "path to the current directory.\n",
	      prog_name);
      exit(1);
   }
   if (NULL == outfilename) {
      l = strlen(cwd) + 1;
      l += strlen(pkg_name) + 1;
      l += strlen(pkg_version) + 1;
      l += strlen(pkg_release) + 1;
      l += strlen(pkg_arch) + 1;
      l += strlen(tar_suffix) + 1;
      outfilename = malloc(l);
      if (NULL == outfilename) {
	 fprintf(stderr,
		 "%s: error: I seem to have run out of memory.\n",
		 prog_name);
	 exit(1);
      }
      sprintf(outfilename, "%s/%s-%s-%s.%s%s", cwd, pkg_name, pkg_version,
	      pkg_release, pkg_arch, tar_suffix);
   }
   
   /* Create working dir */
   snprintf(buffer, BUF_LEN, "rm -rf %s", work_dir);
   do_command(buffer);
   snprintf(buffer, BUF_LEN, "mkdir -p %s", work_dir);
   do_command(buffer);
   
   /* Uncpio the RPM */
   fprintf(stderr, "+ cd %s\n", work_dir);
   status = chdir(work_dir);
   if (0 != status) {
      perror(work_dir);
      exit(1);
   }
   if ('/' == pkg_filename[0]) {
      snprintf(buffer, BUF_LEN, "rpm2cpio %s |%s", pkg_filename, cpio_cmd);
   } else {
      snprintf(buffer, BUF_LEN, "rpm2cpio %s/%s |%s",
	       cwd, pkg_filename, cpio_cmd);
   }
   do_command(buffer);
   
   /* Change ownership and modes */
   snprintf(buffer, BUF_LEN, "chmod -R u+rwX,go+rX .");
   do_command(buffer);
   snprintf(buffer, BUF_LEN, "chown -R root.root .");
   do_command(buffer);
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_OLDFILENAMES, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the package filenames.\n",
	      prog_name);
      exit(1);
   }
   pkg_files = (char **)entry;
   n = entry_count;
#if DEBUG_STUFF
   fprintf(stderr, "Number of files = %d\n", n);
#endif
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_FILEMODES, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the file modes.\n",
	      prog_name);
      exit(1);
   }
   if (entry_count != n) {
      fprintf(stderr,
	      "%s: error: The number of file modes contained in the "
	      "package header seems to be different than the number "
	      "of filenames.\n",
	      prog_name);
      exit(1);
   }
   pkg_modes = (int_16 *)entry;
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_FILEUSERNAME, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the file owners.\n",
	      prog_name);
      exit(1);
   }
   if (entry_count != n) {
      fprintf(stderr,
	      "%s: error: The number of file owners contained in the "
	      "package header seems to be different than the number "
	      "of filenames.\n",
	      prog_name);
      exit(1);
   }
   pkg_owners = (char **)entry;
   
   status = rpmHeaderGetEntry(pkg_hdr, RPMTAG_FILEGROUPNAME, &entry_type,
			      (void **)(&entry), &entry_count);
   if (!status) {
      fprintf(stderr,
	      "%s: error: That's odd; the package header doesn't seem to "
	      "have an entry for the file groups.\n",
	      prog_name);
      exit(1);
   }
   if (entry_count != n) {
      fprintf(stderr,
	      "%s: error: The number of file groups contained in the "
	      "package header seems to be different than the number "
	      "of filenames.\n",
	      prog_name);
      exit(1);
   }
   pkg_groups = (char **)entry;
   
   for (i = 0; i < n; i++) {
      f = pkg_files[i];
      if ('/' == *f) {
	 f++;
      }
      owner = pkg_owners[i];
      group = pkg_groups[i];
      if ((0 != strcmp(owner, "root")) ||
	  (0 != strcmp(group, "root"))) {
	 snprintf(buffer, BUF_LEN, "chown %s.%s %s", owner, group, f);
	 do_command(buffer);
      }
      status = chmod(f, (mode_t)pkg_modes[i]);
      if (0 != status) {
	 perror(f);
      }
   }
   free(pkg_files);
   
   /* Tar */
   if (0 == strcmp(outfilename, "-")) {
      snprintf(buffer, BUF_LEN, "%s", tar_cmd);
   } else {
      snprintf(buffer, BUF_LEN, "%s >%s", tar_cmd, outfilename);
   }
   if (l) {
      free(outfilename);
   }
   do_command(buffer);
   
   /* Remove directory? */
   fprintf(stderr, "+ cd %s\n", cwd);
   status = chdir(cwd);
   if (0 != status) {
      perror(cwd);
      exit(1);
   }
   snprintf(buffer, BUF_LEN, "rm -rf %s", work_dir);
   do_command(buffer);
   
   return(0);
}
